use std::sync::Arc;

use axum::{Extension, Json};
use foxchat::{
    http::ApiError,
    model::{channel::PartialChannel, http::guild::CreateGuildParams, Guild},
    s2s::http::RestEvent,
    FoxError,
};

use crate::{
    app_state::AppState,
    db::{
        channel::create_channel,
        guild::{create_guild, join_guild},
    },
    fed::{post_rest_event, FoxRequestData},
    model::user::User,
};

pub async fn post_guilds(
    Extension(state): Extension<Arc<AppState>>,
    request: FoxRequestData,
    Json(params): Json<CreateGuildParams>,
) -> Result<Json<Guild>, ApiError> {
    let user_id = request.user_id.ok_or(FoxError::MissingUser)?;
    let user = User::get(&state, &request.instance, &user_id).await?;

    let mut tx = state.pool.begin().await?;
    let guild = create_guild(&mut *tx, &user.id, &params.name).await?;
    let channel = create_channel(&mut *tx, &guild.id, "general", None).await?;

    join_guild(&mut *tx, &guild.id, &user.id).await?;
    tx.commit().await?;

    // Send an event to the user's instance that they joined a guild
    post_rest_event(
        state,
        &request.instance,
        RestEvent::GuildJoin {
            guild_id: guild.id.0.clone(),
            user_id: user.remote_user_id.clone(),
        },
    )
    .await
    .ok();

    Ok(Json(Guild {
        id: guild.id.0,
        name: guild.name,
        owner_ids: vec![user.id.0],
        channels: Some(vec![PartialChannel {
            id: channel.id.0,
            name: channel.name,
        }]),
    }))
}
