use base64::prelude::{Engine, BASE64_URL_SAFE};
use chrono::{DateTime, Utc, Duration, NaiveDateTime};
use eyre::Result;
use rsa::pkcs1v15::{SigningKey, VerifyingKey, Signature};
use rsa::sha2::Sha256;
use rsa::signature::{RandomizedSigner, SignatureEncoding, Verifier};
use rsa::{RsaPrivateKey, RsaPublicKey};

use crate::FoxError;

pub fn build_signature(
    private_key: &RsaPrivateKey,
    host: &str,
    request_path: &str,
    content_length: Option<usize>,
    user_id: Option<String>,
) -> (String, DateTime<Utc>) {
    let mut rng = rand::thread_rng();
    let signing_key = SigningKey::<Sha256>::new(private_key.clone());
    let time = Utc::now();

    let plaintext = plaintext_string(time, &host, &request_path, content_length, user_id);
    let signature = signing_key.sign_with_rng(&mut rng, plaintext.as_bytes());

    let str = BASE64_URL_SAFE.encode(signature.to_bytes());

    (str, time)
}

fn plaintext_string(
    time: DateTime<Utc>,
    host: &str,
    request_path: &str,
    content_length: Option<usize>,
    user_id: Option<String>,
) -> String {
    let raw_time = format_date(time);
    let raw_content_length = content_length
        .map(|i| i.to_string())
        .unwrap_or("".to_owned());
    let raw_user_id = user_id.unwrap_or("".into());

    let s = format!(
        "{}:{}:{}:{}:{}",
        raw_time, host, request_path, raw_content_length, raw_user_id
    );

    s
}

pub fn format_date(time: DateTime<Utc>) -> String {
    time.format("%a, %d %b %Y %T GMT").to_string()
}

pub fn parse_date(input: &str) -> Result<DateTime<Utc>, chrono::ParseError> {
    Ok(NaiveDateTime::parse_from_str(input, "%a, %d %b %Y %T GMT")?.and_utc())
}

pub fn verify_signature(
    public_key: &RsaPublicKey,
    encoded_signature: String,
    time: DateTime<Utc>, // from Date header
    host: &str, // from Host header, verify that it's actually your host
    request_path: &str, // from router
    content_length: Option<usize>, // from Content-Length header
    user_id: Option<String>, // from X-Foxchat-User header
) -> Result<bool> {
    let verifying_key = VerifyingKey::<Sha256>::new(public_key.clone());

    let now = Utc::now();
    if (now - Duration::minutes(1)) > time {
        return Err(FoxError::SignatureDateOutOfRange("request was made too long ago").into());
    } else if (now + Duration::minutes(1)) < time {
        return Err(FoxError::SignatureDateOutOfRange("request was made in the future").into());
    }

    let plaintext = plaintext_string(time, host, request_path, content_length, user_id);

    let hash = &BASE64_URL_SAFE.decode(encoded_signature)?;
    let signature = Signature::try_from(hash.as_ref())?;

    verifying_key.verify(plaintext.as_bytes(), &signature)?;
    Ok(true)
}
